import asyncio
import collections
import json
import random

import aiohttp
from aiohttp import WSMessage

_MEM = 'μοκιε'


class MokeiWebSocketClient:
    def __init__(self, url: str):
        self.url = url
        self._ws = None
        self._default_backoff = 1.0
        self._current_backoff = self._default_backoff
        self._max_backoff = 15.0
        self._unsent_messages = collections.deque()
        self._onconnect_handlers = []
        self._ontext_handlers = []
        self._ondisconnect_handlers = []
        self._handlers: dict[str, list] = collections.defaultdict(list)
        self._session: aiohttp.ClientSession | None = None

    def _get_backoff(self):
        backoff = self._current_backoff
        self._current_backoff += self._current_backoff * random.random()
        self._current_backoff = min(self._current_backoff, self._max_backoff)
        return backoff

    def _reset_backoff(self):
        self._current_backoff = self._default_backoff

    async def _onconnect_handler(self, ws):
        await asyncio.gather(*(handler(ws) for handler in self._onconnect_handlers))

    async def _ondisconnect_handler(self, ws):
        await asyncio.gather(*(handler(ws) for handler in self._ondisconnect_handlers))

    async def _ontext_handler(self, ws, msg: str):
        if msg.startswith(_MEM):
            event_data = json.loads(msg[len(_MEM):])
            if 'event' not in event_data or 'data' not in event_data:
                return
            event = event_data['event']
            data = event_data['data']
            await asyncio.gather(*[handler(ws, data) for handler in self._handlers[event]])
        await asyncio.gather(*[handler(ws, msg) for handler in self._ontext_handlers])

    def onconnect(self, handler):
        """Decorator method.

        Decorate an async function which accepts one argument (a mokei.Websocket), and returns None

        Example:

        client = MokeiWebSocketClient('https://someurl.com')

        @client.onconnect
        async def connectionhandler(socket: mokei.WebSocket) -> None:
            logger.info(f'New connection from {socket.request.remote}')
        """
        self._onconnect_handlers.append(handler)
        return handler

    def ondisconnect(self, handler):
        """Decorator method.

        Decorate an async function which accepts one argument (a mokei.Websocket), and returns None

        Example:

        client = MokeiWebSocketClient('https://someurl.com')

        @client.ondisconnect
        async def disconnecthandler(socket: mokei.WebSocket) -> None:
            logger.info(f'Lost connection to {socket.request.remote}')
        """
        self._ondisconnect_handlers.append(handler)
        return handler

    async def connect(self):
        self._session = aiohttp.ClientSession()
        async with self._session as session:
            while True:
                try:
                    async with session.ws_connect(self.url) as ws:
                        self._reset_backoff()
                        self._ws = ws
                        await self._onconnect_handler(ws)
                        await self._send_unsent_messages()
                        async for msg in ws:
                            msg: WSMessage
                            if msg.type == aiohttp.WSMsgType.TEXT:
                                await self._ontext_handler(ws, msg.data)
                            elif msg.type == aiohttp.WSMsgType.ERROR:
                                break
                except aiohttp.ClientConnectorError:
                    pass
                except RuntimeError:
                    raise RuntimeError

                if self._ws:
                    await self._ondisconnect_handler(self._ws)
                self._ws = None
                await asyncio.sleep(self._get_backoff())

    async def close(self):
        await self._session.close()

    async def _send_unsent_messages(self):
        while self._unsent_messages:
            try:
                if not self._ws:
                    break
                await self._ws.send_str(self._unsent_messages[0])
                self._unsent_messages.popleft()
            except ConnectionResetError:
                break

    async def send(self, text: str):
        self._unsent_messages.append(text)
        await self._send_unsent_messages()

    def ontext(self, handler):
        self._ontext_handlers.append(handler)
        return handler

    def on(self, event: str):
        def decorator(fn):
            self._handlers[event].append(fn)
            return fn

        return decorator



import asyncio
import functools
import inspect
import pathlib
import ssl
from typing import Optional, Callable, Any, Awaitable

from aiohttp import web, WSMessage, WSMsgType
from jinja2 import FileSystemLoader
from aiohttp_asgi import ASGIResource

from .config import Config
from .exceptions import MokeiConfigError
from .request import Request
from .websocket import MokeiWebSocket, MokeiWebSocketRoute
from .logging import getLogger

logger = getLogger(__name__)

TemplateContext = dict[str, Any]


class Mokei:
    def __init__(
            self,
            config: Optional[Config] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        config = config or Config()
        self.config = config
        self._loop = loop or asyncio.new_event_loop()
        self._web_app: Optional[web.Application] = None
        self._asgi_resouces: list[ASGIResource] = []
        self._routes = web.RouteTableDef()
        self.view = self._routes.view
        self._handlers = {}
        self._ssl_context = None
        if config.template_dir and isinstance(config.template_dir, str):
            config.template_dir = pathlib.Path(config.template_dir)
        self._template_dir: Optional[pathlib.Path] = config.template_dir
        self._static_dirs: dict[str, pathlib.Path] = {}
        # list of zero-arg async functions that return None
        self._background_tasks: list[Callable[[], Awaitable[None]]] = []

    def _get_default_template_dir(self) -> Optional[pathlib.Path]:
        for dirname in ['template', 'templates']:
            templatedir = pathlib.Path(dirname)
            if templatedir.exists():
                return templatedir

    def _get_default_static_dir(self) -> Optional[pathlib.Path]:
        for dirname in ['static', 'public']:
            staticdir = pathlib.Path(dirname)
            if staticdir.exists():
                self.set_static_dir(f'/{dirname}', staticdir)

    def redirect(self, target):
        if isinstance(target, str):
            if target.startswith('/'):
                raise web.HTTPFound(target)
            else:
                if target in self._handlers:
                    raise web.HTTPFound(self._handlers[target])
        elif isinstance(target, Callable):
            if target.__name__ in self._handlers:
                raise web.HTTPFound(self._handlers[target.__name__])

    @staticmethod
    def _get_normalized_handler(raw_handler):
        if not asyncio.iscoroutinefunction(raw_handler):
            raise TypeError('handler must be an async function')

        if getattr(raw_handler, '_is_mokei_normalized', False):
            # this handler has been normalized already - return as is
            return raw_handler

        sig = inspect.signature(raw_handler)
        params = sig.parameters

        def first_param_is_request() -> bool:
            try:
                first_param = next(iter(sig.parameters.items()))
                # noinspection PyProtectedMember,PyUnresolvedReferences
                if (
                        first_param[0] == 'request' and first_param[1].annotation is inspect._empty
                        or first_param[1].annotation is Request):
                    return True
            except StopIteration:
                pass
            return False

        async def normalized_handler(r: Request):
            kwargs = {key: r.match_info[key] for key in r.match_info if key in params}
            return await raw_handler(r, **kwargs)

        async def normalized_handler_with_request(r: Request):
            kwargs = {key: r.match_info[key] for key in r.match_info if key in params}
            return await raw_handler(**kwargs)

        if first_param_is_request():
            handler = normalized_handler
        else:
            handler = normalized_handler_with_request

        handler._is_mokei_normalized = True
        return handler

    def template(self, template_name: str):
        """Decorator method to decorate routes which use templates."""
        if not self.config.use_templates:
            raise MokeiConfigError('Set Config.use_templates to True to use templates')
        from aiohttp_jinja2 import template

        def decorator(handler):
            normalized_handler = self._get_normalized_handler(handler)
            return template(template_name)(normalized_handler)

        return decorator

    def _route_handler(self, path: str, http_method: str):
        """Generic decorator method for handling any http method
        Do not call this method directly, but use partialmethod to create methods for handling specific http methods
        """

        def decorator(handler):
            if not asyncio.iscoroutinefunction(handler):
                raise TypeError('Handler must be a async function')
            normalized_handler = self._get_normalized_handler(handler)
            getattr(self._routes, http_method)(path)(normalized_handler)
            self._handlers[handler.__name__] = path
            return handler

        return decorator

    get = functools.partialmethod(_route_handler, http_method='get')
    post = functools.partialmethod(_route_handler, http_method='post')
    head = functools.partialmethod(_route_handler, http_method='head')
    patch = functools.partialmethod(_route_handler, http_method='patch')
    delete = functools.partialmethod(_route_handler, http_method='delete')
    put = functools.partialmethod(_route_handler, http_method='put')

    def background_task(self, fn):
        self._background_tasks.append(fn)
        return fn

    def set_static_dir(self, prefix: str, dir_path: str | pathlib.Path) -> 'Mokei':
        if prefix not in self._static_dirs:
            self._static_dirs[prefix] = dir_path
            self._routes.static(prefix, dir_path)
        return self

    def set_template_dir(self, template_dir: str | pathlib.Path) -> 'Mokei':
        self.config.template_dir = pathlib.Path(template_dir)
        return self

    def load_cert(self, certfile, keyfile=None, password=None) -> 'Mokei':
        if self._ssl_context is None:
            self._ssl_context = ssl.create_default_context(
                ssl.Purpose.CLIENT_AUTH
            )
        self._ssl_context.load_cert_chain(certfile, keyfile, password)
        return self

    def register_asgi_app(self, asgi_app, path: str) -> None:
        self._asgi_resouces = ASGIResource(asgi_app, root_path=path)

    def run(
            self,
            config: Optional[Config] = None,
            **kwargs,
    ) -> None:
        """kwargs will be passed directly to Config
         See Config definition more information for acceptable kwarg names and types
        """
        if config is not None:
            self.config.__dict__.update(config.__dict__)
        for key, value in kwargs.items():
            setattr(self.config, key, value)
        # create aiohttp web application
        self._web_app = web.Application(middlewares=[])
        self._web_app.add_routes(self._routes)
        self._web_app['state'] = {}
        self._web_app.middlewares.extend(self.config.middlewares)

        # set up aiohttp_jinja2 templates, if necessary
        if self.config.use_templates:
            if self._template_dir is None:
                self._template_dir = self._get_default_template_dir()
            if self._template_dir is not None:
                from aiohttp_jinja2 import setup as template_setup
                template_setup(self._web_app, loader=FileSystemLoader(self._template_dir))

        # setup aiohttp_swagger documentation
        if self.config.use_swagger:
            from aiohttp_swagger import setup_swagger
            setup_swagger(self._web_app)

        # load certificates, if available
        if self.config.certfile is not None:
            self.load_cert(self.config.certfile, self.config.keyfile, self.config.password)

        # create background tasks
        for background_task in self._background_tasks:
            self._loop.create_task(background_task())

        # register any asgi resources
        for asgi_resource in self._asgi_resouces:
            self._web_app.router.register_resource(asgi_resource)

        # run the application
        web.run_app(self._web_app, host=self.config.host, port=self.config.port,
                    ssl_context=self._ssl_context, loop=self._loop)

    def websocketroute(self, path: str) -> MokeiWebSocketRoute:
        socket_route = MokeiWebSocketRoute(path)

        async def ws(request: Request) -> MokeiWebSocket:
            return await self._handle_websocket(request, socket_route)

        self._routes.get(path)(ws)
        return socket_route

    async def _handle_websocket(self, request: Request, socket_route: MokeiWebSocketRoute) -> MokeiWebSocket:
        """ Called when a new websocket connection is established from a client
        """
        ws = MokeiWebSocket(request, socket_route)
        try:
            await ws.prepare(request)
        except ConnectionResetError:
            return ws
        socket_route.sockets.add(ws)
        # noinspection PyProtectedMember
        await socket_route._onconnect_handler(ws)
        async for msg in ws:
            msg: WSMessage
            if msg.type is WSMsgType.TEXT:
                # noinspection PyProtectedMember
                await socket_route._ontext_handler(ws, msg.data)
            elif msg.type is WSMsgType.BINARY:
                # noinspection PyProtectedMember
                await socket_route._onbinary_handler(ws, msg.data)

        socket_route.sockets.remove(ws)
        # noinspection PyProtectedMember
        await self._loop.create_task(socket_route._ondisconnect_handler(ws))
        return ws



import asyncio
import collections
import json
from typing import Callable, Awaitable, Optional, Iterable
import uuid

from aiohttp import web

from .datatypes import JsonDict
from .request import Request
from .logging import getLogger

logger = getLogger(__name__)

# MokeiEventMarket, a marker prepended to json data in raw message when sending/receiving events (rather than text)
_MEM = 'μοκιε'


class MokeiWebSocket(web.WebSocketResponse):
    def __init__(self, request: Request, route: 'MokeiWebSocketRoute', *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.id = uuid.uuid4()
        self.request = request
        self._route = route

    def __repr__(self) -> str:
        return f'<MokeiWebSocket {self.request.remote} {self.id}>'

    def __bool__(self) -> bool:
        # Ensure that "if websocket:" works properly.
        # For some reason bool(inst_of_superclass) evaluates to False
        return True

    async def send_text(self, message: str) -> None:
        await self._route.send_text(message, self)

    async def send_event(self, event: str, data: JsonDict) -> None:
        await self._route.send_event(event, data, self)


OnConnectHandler = Callable[[MokeiWebSocket], Awaitable[None]]
OnDisconnectHandler = Callable[[MokeiWebSocket], Awaitable[None]]
OnEventHandler = Callable[[MokeiWebSocket, JsonDict], Awaitable[None]]
OnTextHandler = Callable[[MokeiWebSocket, str], Awaitable[None]]
OnBinaryHandler = Callable[[MokeiWebSocket, bytes], Awaitable[None]]


class MokeiWebSocketRoute:
    def __init__(self, path) -> None:
        self.path = path
        self._onconnect_handlers: list[OnConnectHandler] = []
        self._ondisconnect_handlers: list[OnDisconnectHandler] = []
        self._ontext_handlers: list[OnTextHandler] = []
        self._onbinary_handlers: list[OnBinaryHandler] = []
        self._onevent_handlers: dict[str, list[OnEventHandler]] = collections.defaultdict(list)
        self.sockets: set[MokeiWebSocket] = set()

    async def _onconnect_handler(self, ws: MokeiWebSocket) -> None:
        """Internal method called when a new websocket connection is received
        This method calls all handlers registered by ws.onconnect
        in the order that they were registered.
        """
        await asyncio.gather(*(handler(ws) for handler in self._onconnect_handlers))

    async def _ondisconnect_handler(self, ws: MokeiWebSocket) -> None:
        """Internal method called when a websocket disconnects
        This method calls all handlers registered by ws.disonconnect
        in the order that they were registered.
        """
        await asyncio.gather(*(handler(ws) for handler in self._ondisconnect_handlers))

    async def _ontext_handler(self, ws: MokeiWebSocket, message: str) -> None:
        """Internal method called when a websocket receives any text
        This method checks if the text is a Mokei Event (i.e. starts with _MEM)
        If it is an event, all self._onenvent_handlers are called in order
        If not, then all self._ontext_handlers are called in order
        """
        logger.debug(message)
        if message.startswith(_MEM):
            event_dict = json.loads(message[len(_MEM):])
            event = event_dict.get('event')
            data_dict = event_dict.get('data')
            if not event:
                return
            handlers = self._onevent_handlers.get(event)
            await asyncio.gather(*(handler(ws, data_dict) for handler in handlers))
        else:
            await asyncio.gather(*(handler(ws, message) for handler in self._ontext_handlers))

    async def _onbinary_handler(self, ws: MokeiWebSocket, message: bytes) -> None:
        """Internal method called when a websocket receives any binary
        """
        await asyncio.gather(*(handler(ws, message) for handler in self._onbinary_handlers))

    def onconnect(self, handler):
        """Decorator for async functions to be run when a new websocket connection is received

        @yourwebsocketroute.onconnect
        async def send_welcome_message(websocket: Websocket):
            logger.info(f'New connection from {websocket.remote}'
            await websocket.send_text('Welcome!')
        """
        self._onconnect_handlers.append(handler)
        return handler

    def ondisconnect(self, handler):
        """Decorator for async functions to be run when a websocket connection is closed

        @yourwebsocketroute.ondisconnect
        async def send_welcome_message(websocket: Websocket):
            logger.info('Websocket from %s disconnected', websocket.remote)
        """
        self._ondisconnect_handlers.append(handler)
        return handler

    def on(self, event: str) -> Callable[[OnEventHandler], OnEventHandler]:
        """Decorator for mokei events

        @yourwebsocketroute.on('my_event')
        async def log_event(websocket: Websocket, data: JsonData):
            logger.info('Received my_event')
            logger.info(data)
        """

        def decorator(handler: OnEventHandler) -> OnEventHandler:
            self._onevent_handlers[event].append(handler)
            return handler

        return decorator

    def ontext(self, handler: OnTextHandler) -> OnTextHandler:
        self._ontext_handlers.append(handler)
        return handler

    def onbinary(self, handler: OnBinaryHandler) -> OnBinaryHandler:
        self._onbinary_handlers.append(handler)
        return handler

    async def send_text(self, message: str, *target: MokeiWebSocket,
                        exclude: Optional[MokeiWebSocket | Iterable[MokeiWebSocket]] = None) -> None:
        # handle cases where exclude is None or a single MokeiWebSocket
        exclude = exclude or ()

        if isinstance(exclude, MokeiWebSocket):
            # harmonize arg "exclude" always to be Iterable[MokeiWebSocket]
            exclude = (exclude,)

        # create a list of sockets to be removed (for failure) post-send
        remove_sockets = list()

        # create a set of recipient sockets (target is just an Iterable[WebSocket] at this point)
        if target:
            recipient_sockets = {target_socket for target_socket in target}
        else:
            # target all sockets in self by default, unless specifically provided in args
            recipient_sockets = {target_socket for target_socket in self.sockets}

        # remove from recipient_sockets any sockets listed in exclude (affects this one event only)
        for socket_to_remove in exclude:
            if socket_to_remove in recipient_sockets:
                recipient_sockets.remove(socket_to_remove)

        async def send_to_single_ws(_message: str, _ws: MokeiWebSocket):
            """Send text to a single websocket
            """
            try:
                await _ws.send_str(_message)
            except ConnectionResetError:
                # unexpected disconnect from remote side
                remove_sockets.append(_ws)
                if _ws in self.sockets:
                    self.sockets.remove(_ws)

        # send the event
        await asyncio.gather(*(send_to_single_ws(message, recipient_socket) for recipient_socket in recipient_sockets))

        # remove any failed sockets from this route
        for socket_to_remove in remove_sockets:
            if socket_to_remove in self.sockets:
                self.sockets.remove(socket_to_remove)

    async def send_event(self, event: str, data: JsonDict, *target: MokeiWebSocket,
                         exclude: Optional[MokeiWebSocket | Iterable[MokeiWebSocket]] = None) -> None:

        message = _MEM + json.dumps({'event': event, 'data': data})

        await self.send_text(message, *target, exclude=exclude)

    def __repr__(self):
        return f'<WebSocketRoute {self.path}>'



from aiohttp import web


@web.middleware
async def mokei_resp_type_middleware(request, handler):
    resp = await handler(request)
    if isinstance(resp, tuple) and len(resp) == 2 and isinstance(resp[1], int):
        status = resp[1]
        resp = resp[0]
    else:
        status = 200
    if isinstance(resp, dict):
        return web.json_response(resp, status=status)
    if isinstance(resp, str):
        return web.Response(body=resp, status=status)
    return resp



from typing import Optional

import aiohttp
from aiohttp.client import ClientSession
from yarl import URL

from .datatypes import JsonDict


class MokeiClient:
    def __init__(self, verify_ssl: bool = True,
                 headers: Optional[dict[str, str]] = None,
                 cookies: Optional[dict[str, str]] = None):
        self._session: ClientSession = ClientSession()
        self.verify_ssl = verify_ssl
        self.headers = headers or {}
        self.cookies = cookies or {}

    async def start_new_session(self, keep_headers: bool = False, keep_cookies: bool = False) -> None:
        if self._session:
            await self._session.close()
        if not keep_headers:
            self.headers.clear()
        if not keep_cookies:
            self.cookies.clear()

    async def post_form_data(self, url: str | URL, data: dict) -> tuple[str, int]:
        form_data = aiohttp.FormData()
        for key, value in data.items():
            form_data.add_field(key, value)
        async with self._session.post(url, data, headers=self.headers, verify_ssl=self.verify_ssl) as response:
            return await response.text(), response.status

    async def post_json(self, url: str | URL, jsondict: JsonDict) -> tuple[str, int]:
        async with self._session.post(url, json=jsondict, headers=self.headers, verify_ssl=self.verify_ssl) as response:
            return await response.text(), response.status

    async def get(self, url: str | URL) -> tuple[str, int]:
        async with self._session.get(url, headers=self.headers, verify_ssl=self.verify_ssl) as response:
            return await response.text(), response.status

    def set_header(self, header_name: str, header_value: str) -> None:
        self.headers[header_name] = header_value

    def set_cookie(self, cookie_name: str, cookie_value: str) -> None:
        self.cookies[cookie_name] = cookie_value

    def set_headers(self, headers: dict[str, str]) -> None:
        for header, value in headers.items():
            self.set_header(header, value)

    def set_cookies(self, cookies: dict[str, str]) -> None:
        for cookie, value in cookies.items():
            self.set_cookie(cookie, value)



import sys as _sys
import time as _time
import logging as _logging
from logging import *

_logging.Formatter.converter = _time.gmtime
_logging.basicConfig(
    stream=_sys.stdout,
    level=INFO,
    format='%(asctime)s %(levelname)s - %(message)s'
)

_aiohttp_loggers = (
    'aiohttp.access',
    'aiohttp',
    'aiohttp.client',
    'aiohttp.internal',
    'aiohttp.server',
    'aiohttp.web',
    'aiohttp.websocket',
)

for logger in _aiohttp_loggers:
    getLogger(logger).setLevel(ERROR)



"""mokei

By deuxglaces
"""

from .config import Config
from .exceptions import MokeiException, MokeiConfigError
from .mokei import Mokei, TemplateContext
from .request import Request
from .websocket import MokeiWebSocket, MokeiWebSocketRoute
from .client import MokeiClient
from .wsclient import MokeiWebSocketClient

__version__ = '0.2.4'
